function CCMF_ML(FD, FD_seeds, FD_out)

img = double(metaImageRead(FD));
img_hdr = metaImageInfo(FD);
image_seeds = metaImageRead(FD_seeds);

% define the label information
nlab = 3; 

%%%%%%%%%%%%%%% image size
[rows, cols, slices] = size(img);

%%%%%%%%%%%%%%% image normalization
img_normalized = mat2gray(img);

%%%%%%%%%%%%%% image filtering 
h = fspecial('gaussian',5,1);
img_smooth = imfilter(img_normalized,h,'replicate');
img_normalized = img_smooth;

%%%%%%%%%%%%%% gradients and penalty %%%%%%%%%%
ALPHA = zeros (rows,cols,slices);
for  i = 1 : slices
    [dx,dy] = gradient(255*img_smooth(:,:,i));
    grad = sqrt(dx.^2+dy.^2);
    max_grad = max(grad(:));
    min_grad = min(grad(:));
    grad = (grad-min_grad)./(max_grad-min_grad);

    ALPHA(:,:,i) = 2.5 + 10*exp(-100*grad); % for GPU  
end

%%%%%%%%%%%%%% data terms
big_num = 1e4;
scale = 700;
samples = 0:1:scale;
Width = 2;

IND = round(scale*img_normalized)+1;
CT = zeros (rows,cols,slices,nlab);

%%%%%%%%%%%%% pdf
%right lung
Ivals = img_normalized(find(image_seeds == 1));
Ivals = Ivals*scale;
pdf_right_lung = ksdensity(Ivals,samples,'Width',Width);
look_right_lung = -log(pdf_right_lung./length(Ivals)+eps);

%left lung
Ivals = img_normalized(find(image_seeds == 2));
Ivals = Ivals*scale;
pdf_left_lung = ksdensity(Ivals,samples,'Width',Width);
look_left_lung = -log(pdf_left_lung./length(Ivals)+eps);

%background
Ivals = img_normalized(find(image_seeds == 3));
Ivals = Ivals*scale;
pdf_bg= ksdensity(Ivals,samples,'Width',Width);
look_bg = -log(pdf_bg./length(Ivals)+eps);

%we can assume the firt and last 20 slices in L/R, S/I are background
side_thickness = 20;

%%%%%%%%%%% source and sink flows
for i = 1: slices  
    % label 1 for slice i
    tem = look_right_lung(IND(:,:,i));
    tem (image_seeds(:,:,i) == 1) = 0;
    tem (image_seeds(:,:,i) == 2) = big_num;
    tem (image_seeds(:,:,i) == 3) = big_num;
    tem(:,[1:side_thickness,end-side_thickness:end]) = big_num;
    tem([1:side_thickness,end-side_thickness:end],:) = big_num;
    CT(:,:,i,1) = tem;
    
    % label 2 for slice i
    tem = look_left_lung(IND(:,:,i));
    tem (image_seeds(:,:,i) == 1) = big_num;
    tem (image_seeds(:,:,i) == 2) = 0;
    tem (image_seeds(:,:,i) == 3) = big_num;
    tem(:,[1:side_thickness,end-side_thickness:end]) = big_num;
    tem([1:side_thickness,end-side_thickness:end],:) = big_num;
    CT(:,:,i,2) = tem;

    % label 3 for slice i
    tem = look_bg(IND(:,:,i));
    tem (image_seeds(:,:,i) == 1) = big_num;
    tem (image_seeds(:,:,i) == 2) = big_num;
    tem (image_seeds(:,:,i) == 3) = 0;
    tem(:,[1:side_thickness,end-side_thickness:end]) = 0;
    tem([1:side_thickness,end-side_thickness:end],:) = 0;
    CT(:,:,i,3) = tem;
end
CT = 0.25*CT;

%%%%%%%%%%%%%%%%%%%%%% segmentation
% set up the input parameter vector
iterNum = 800;
err_bound = 1e-6;
cc = 0.25;
steps = 0.15;
beta = 0.75;
gamma1=1.0;
gamma2=1.3;
varParas = [rows; cols; slices; iterNum; err_bound; cc; steps; nlab; beta; gamma1; gamma2];
%                para 0,1,2 - rows, cols, slices
%                para 3 - the maximum number of iterations
%                para 4 - the error bound for convergence
%                para 5 - cc for the step-size of augmented Lagrangian method
%                para 6 - the step-size for the graident-projection of the spatial flow fields p(x,i)
%                para 7 - the total number of labels
%                para 8 - similarity weight
%                para 9 - left to right lung volume ratio
%                para 10 -ratio weight
penalty = ALPHA;
% ----------------------------------------------------------------------
%  Use the function CCMF_ML_GPU to run the algorithm on GPU
% ----------------------------------------------------------------------

[u, erriter, num, timet] = CCMF_ML_GPU(single(penalty), single(CT), single(varParas));
u = cc*u;

for i = 1 : slices
    [um,I(:,:,i)] = max(permute(u(:,:,i,:),[1 2 4 3]), [], 3);
end

% reconstructing the image with the computed label-indicator function I(x)
uu = I;
uu(I == 1) = 1;
uu(I == 2) = 2;
uu(I == 3) = 0;

% metaImageWrite(uu,[path labelFileName], img_hdr);
metaImageWrite(uu, FD_out, img_hdr);
end
