#include <stdio.h>

#define BLOCK_SIZE 8
#define BLOCK2D_SIZE 32
#define MAX(a,b) ( a > b ? a : b )
#define MIN(a,b) ( a <= b ? a : b )
#define SIGN(x) ( x >= 0.0 ? 1.0 : -1.0 )
#define ABS(x) ( (x) > 0.0 ? x : -(x) )
#define SQR(x) __fmul_rz((x), (x))

static __global__ void krnl_1(float *pfdv, float *pfpt, float *pfps, 
                            float *pfu, float *pfgk, float cc, 
                            int iNx, int iNy, int iLab, int iNz, int SZF, int szImg,
                            int blocksInY, double invBlocksInY, float *pfqq, float *pfpw)
{
   //int blockIdxz = __double2uint_rd(blockIdx.y * invBlocksInY);
   //int blockIdxy = blockIdx.y - __umul24(blockIdxz,blocksInY);
   int blockIdxz = __float2uint_rd(__fdiv_rd(blockIdx.y,blocksInY)); 
   int blockIdxy = blockIdx.y - __umul24(blockIdxz,blocksInY);
   int idx   = __mul24(blockIdx.x,blockDim.x)+threadIdx.x;
   int idy   = __mul24(blockIdxy,blockDim.y)+threadIdx.y;
   int idz   = __mul24(blockIdxz,blockDim.z)+threadIdx.z;
   int index3d, index4d, idlab;

   if( idx<iNy && idy<iNx && idz<iNz)
   {
        for (idlab = 0; idlab <iLab; idlab++)
        {
            index3d     = idx +  __umul24(idy, iNy) +  __umul24(idz, SZF);           

            //index4d     = idx +  __umul24(idy, iNy) +  __umul24(idz, SZF) + __umul24(idlab, szImg);
            index4d     = idx + idy*iNy + idz*SZF + idlab*szImg;
            /* index4d          = idx + idy*iNy + (idz + 0)*iNx*iNy + idlab*iNx*iNy*iLab;  */
            /* index4d - SZF    = idx + idy*iNy + (idz - 1)*iNx*iNy + idlab*iNx*iNy*iLab;  */
            /* index4d + SZF    = idx + idy*iNy + (idz + 1)*iNx*iNy + idlab*iNx*iNy*iLab;  */
            
            /* pfgk[index4d] = __fadd_rz(__fadd_rz(__fadd_rz(pfdv[index4d],-pfps[index3d]), pfpt[index4d]), -pfu[index4d]);*/

            if(idz == 0)
                pfgk[index4d] = __fadd_rz(__fadd_rz(__fadd_rz(__fadd_rz(__fadd_rz(pfdv[index4d], -pfps[index3d]), pfpt[index4d]), -pfu[index4d]), pfqq[index4d]), pfpw[idz*iLab + idlab]);
            else if((idz > 0) && (idz < (iNz - 2)))
                pfgk[index4d] = __fadd_rz(__fadd_rz(__fadd_rz(__fadd_rz(__fadd_rz(__fadd_rz(pfdv[index4d], -pfps[index3d]), pfpt[index4d]), -pfu[index4d]), pfqq[index4d]), -pfqq[index4d - SZF]), pfpw[idz*iLab + idlab]);
            else if(idz == (iNz - 2))
                pfgk[index4d] = __fadd_rz(__fadd_rz(__fadd_rz(__fadd_rz(__fadd_rz(__fadd_rz(pfdv[index4d], -pfps[index3d]), pfpt[index4d]), -pfu[index4d]), pfqq[index4d]), -pfqq[index4d - SZF]), pfpw[idz*iLab + idlab]);
            else if(idz == (iNz - 1))
                pfgk[index4d] = __fadd_rz(__fadd_rz(__fadd_rz(__fadd_rz(__fadd_rz(pfdv[index4d], -pfps[index3d]), pfpt[index4d]), -pfu[index4d]), -pfqq[index4d - SZF]), pfpw[idz*iLab + idlab]);
            else
                ;
        }
    }
}


static __global__ void krnl_2(float *pfbx, float *pfgk, float steps, 
                            int iNx, int iNy, int iLab, int iNz, int SZF, int szImg, 
                            int blocksInY, double invBlocksInY)
{
    //int blockIdxz = __double2uint_rd(blockIdx.y * invBlocksInY);
    //int blockIdxy = blockIdx.y - __umul24(blockIdxz,blocksInY);
    int blockIdxz = __float2uint_rd(__fdiv_rd(blockIdx.y,blocksInY)); 
    int blockIdxy = blockIdx.y - __umul24(blockIdxz,blocksInY);
    int idx   = __mul24(blockIdx.x,blockDim.x)+threadIdx.x;
    int idy   = __mul24(blockIdxy,blockDim.y)+threadIdx.y;
    int idz   = __mul24(blockIdxz,blockDim.z)+threadIdx.z;
    int index4d, idlab;

    if( idx<iNy && idy<(iNx-1) && idz<iNz )
    {
        for (idlab = 0; idlab <iLab; idlab++)
        {
            /* index3d     = idx + __umul24(idy+1, iNy) + __umul24(idz, SZF);           */
            //index4d     = idx + __umul24(idy+1, iNy) + __umul24(idz, SZF) + __umul24(idlab, szImg);
            index4d     = idx + (idy+1)*iNy + idz*SZF + idlab*szImg;
            pfbx[index4d] = __fadd_rz(__fmul_rz(steps,__fadd_rz(pfgk[index4d], -pfgk[index4d-iNy])),pfbx[index4d]);
        }
    }
}

static __global__ void krnl_3(float *pfby, float *pfgk, float steps, 
                            int iNx, int iNy, int iLab, int iNz, int SZF, int szImg, 
                            int blocksInY, double invBlocksInY)
{
    //int blockIdxz = __double2uint_rd(blockIdx.y * invBlocksInY);
    //int blockIdxy = blockIdx.y - __umul24(blockIdxz,blocksInY);
    int blockIdxz = __float2uint_rd(__fdiv_rd(blockIdx.y,blocksInY)); 
    int blockIdxy = blockIdx.y - __umul24(blockIdxz,blocksInY);
    int idx   = __mul24(blockIdx.x,blockDim.x)+threadIdx.x;
    int idy   = __mul24(blockIdxy,blockDim.y)+threadIdx.y;
    int idz   = __mul24(blockIdxz,blockDim.z)+threadIdx.z;
    int index4d, idlab;
    
    if( idx<(iNy-1) && idy<iNx && idz<iNz)
    {
        for (idlab = 0; idlab <iLab; idlab++)
        {
            /* index3d     = idx + 1 + __umul24(idy, iNy) + __umul24(idz, SZF);           */

            //index4d     = idx + 1 + __umul24(idy, iNy) + __umul24(idz, SZF) + __umul24(idlab, szImg);
            index4d       = idx + 1 + idy*iNy + idz*SZF + idlab*szImg;
            pfby[index4d] = __fadd_rz(__fmul_rz(steps, __fadd_rz(pfgk[index4d], -pfgk[index4d-1])), pfby[index4d]);
        }
    }
}

static __global__ void krnl_4(float *pfbx, float *pfby, float *pfgk, 
                            float *pfpenalty, int iNx, int iNy, int iLab, int iNz, 
                            int SZF, int szImg, int blocksInY, double invBlocksInY)
{
    //int blockIdxz = __double2uint_rd(blockIdx.y * invBlocksInY);
    //int blockIdxy = blockIdx.y - __umul24(blockIdxz,blocksInY);
    int blockIdxz = __float2uint_rd(__fdiv_rd(blockIdx.y,blocksInY)); 
    int blockIdxy = blockIdx.y - __umul24(blockIdxz,blocksInY);
    int idx   = __mul24(blockIdx.x,blockDim.x)+threadIdx.x;
    int idy   = __mul24(blockIdxy,blockDim.y)+threadIdx.y;
    int idz   = __mul24(blockIdxz,blockDim.z)+threadIdx.z;
    float fpt;
    int index3d, index4d, idlab;

    if( idx<iNy && idy<iNx && idz<iNz)
    {
        for (idlab = 0; idlab <iLab; idlab++)
        {
            index3d     = idx + __umul24(idy, iNy) + __umul24(idz, SZF); 
            //index4d     = idx + __umul24(idy, iNy) + __umul24(idz, SZF) + __umul24(idlab, szImg);
            index4d     = idx + idy*iNy + idz*SZF + idlab*szImg;

            fpt = __fsqrt_rz(__fmul_rz(__fadd_rz(__fadd_rz(__fadd_rz(SQR(pfbx[index4d+iNy]), SQR(pfby[index4d+1])),SQR(pfbx[index4d])), SQR(pfby[index4d])),0.5));

            if (fpt > pfpenalty[index3d])
                pfgk[index4d] = __fdividef(pfpenalty[index3d], fpt);

            else
                pfgk[index4d] = 1;
        }
    }
}

static __global__ void krnl_5(float *pfbx, float *pfgk, int iNx, int iNy, 
                            int iLab, int iNz, int SZF, int szImg, 
                            int blocksInY, double invBlocksInY)
{
    //int blockIdxz = __double2uint_rd(blockIdx.y * invBlocksInY);
    //int blockIdxy = blockIdx.y - __umul24(blockIdxz,blocksInY);
    int blockIdxz = __float2uint_rd(__fdiv_rd(blockIdx.y,blocksInY)); 
    int blockIdxy = blockIdx.y - __umul24(blockIdxz,blocksInY);
    int idx   = __mul24(blockIdx.x,blockDim.x)+threadIdx.x;
    int idy   = __mul24(blockIdxy,blockDim.y)+threadIdx.y;
    int idz   = __mul24(blockIdxz,blockDim.z)+threadIdx.z;
    int index4d, idlab;
    
    if( idx<iNy && idy<(iNx-1) && idz<iNz)
    {
        for (idlab = 0; idlab <iLab; idlab++)
        {
            /* index3d     = idx + __umul24(idy+1, iNy) + __umul24(idz, SZF);           */
            //index4d     = idx + __umul24(idy+1, iNy) + __umul24(idz, SZF) + __umul24(idlab, szImg);
            index4d     = idx + (idy+1)*iNy + idz*SZF + idlab*szImg;
            pfbx[index4d] = __fmul_rz(__fmul_rz(__fadd_rz(pfgk[index4d], pfgk[index4d-iNy]),0.5), pfbx[index4d]);
        }
    }
}

static __global__ void krnl_6(float *pfby, float *pfgk, int iNx, 
                            int iNy, int iLab, int iNz, int SZF, int szImg, 
                            int blocksInY, double invBlocksInY)
{
    //int blockIdxz = __double2uint_rd(blockIdx.y * invBlocksInY);
    //int blockIdxy = blockIdx.y - __umul24(blockIdxz,blocksInY);
    int blockIdxz = __float2uint_rd(__fdiv_rd(blockIdx.y,blocksInY)); 
    int blockIdxy = blockIdx.y - __umul24(blockIdxz,blocksInY);
    int idx   = __mul24(blockIdx.x,blockDim.x)+threadIdx.x;
    int idy   = __mul24(blockIdxy,blockDim.y)+threadIdx.y;
    int idz   = __mul24(blockIdxz,blockDim.z)+threadIdx.z;
    int index4d, idlab;
    
    if( idx<(iNy-1) && idy<iNx && idz<iNz)
    {
        for (idlab = 0; idlab <iLab; idlab++)
        {
            /* index3d     = idx + 1 + __umul24(idy, iNy) + __umul24(idz, SZF);           */
            //index4d     = idx + 1 + __umul24(idy, iNy) + __umul24(idz, SZF) + __umul24(idlab, szImg);
            index4d       = idx + 1 + idy*iNy + idz*SZF + idlab*szImg;
            pfby[index4d] = __fmul_rz(__fmul_rz(__fadd_rz(pfgk[index4d], pfgk[index4d-1]), 0.5), pfby[index4d]);
        }
    }
}


static __global__ void krnl_7(float *pfbx, float *pfby, float *pfdv,
                            float *pfps, float *pfpt, float *pfu, 
                            float *pfCt, float *pfgk, float cc,
                            int iNx, int iNy, int iLab, int iNz, int SZF, int szImg, 
                            int blocksInY, double invBlocksInY, float beta, float *pfqq, float *pfpw)
{
   //int blockIdxz = __double2uint_rd(blockIdx.y * invBlocksInY);
   //int blockIdxy = blockIdx.y - __umul24(blockIdxz,blocksInY);
   int blockIdxz = __float2uint_rd(__fdiv_rd(blockIdx.y,blocksInY)); 
   int blockIdxy = blockIdx.y - __umul24(blockIdxz,blocksInY);
   int idx   = __mul24(blockIdx.x,blockDim.x)+threadIdx.x;
   int idy   = __mul24(blockIdxy,blockDim.y)+threadIdx.y;
   int idz   = __mul24(blockIdxz,blockDim.z)+threadIdx.z;
   int index3d, index4d, idlab;
   float fpt1, fpt2;

    if( idx<iNy && idy<iNx && idz<iNz)
    {
        for (idlab = 0; idlab <iLab; idlab++)
        {
            index3d     = idx +  __umul24(idy, iNy) +  __umul24(idz, SZF);           
            //index4d     = idx +  __umul24(idy, iNy) +  __umul24(idz, SZF) + __umul24(idlab, szImg);   
            index4d     = idx + idy*iNy + idz*SZF + idlab*szImg;

            /* index4d          = idx + idy*iNy + (idz + 0)*iNx*iNy + idlab*iNx*iNy*iLab;  */
            /* index4d - SZF    = idx + idy*iNy + (idz - 1)*iNx*iNy + idlab*iNx*iNy*iLab;  */
            /* index4d + SZF    = idx + idy*iNy + (idz + 1)*iNx*iNy + idlab*iNx*iNy*iLab;  */

            fpt1 = __fadd_rz(__fadd_rz(__fadd_rz(pfbx[index4d + iNy], -pfbx[index4d]), pfby[index4d + 1]), -pfby[index4d]);

            if(idz == 0)
                fpt2 = __fadd_rz(__fadd_rz(__fadd_rz(__fadd_rz(pfps[index3d], pfu[index4d]), -fpt1), -pfqq[index4d]), -pfpw[idz*iLab + idlab]);
            else if((idz > 0) && (idz < iNz - 2))
                fpt2 = __fadd_rz(__fadd_rz(__fadd_rz(__fadd_rz(__fadd_rz(pfps[index3d], pfu[index4d]), -fpt1), -pfqq[index4d]), pfqq[index4d - SZF]), -pfpw[idz*iLab + idlab]);
            else if(idz == iNz - 2)
                fpt2 = __fadd_rz(__fadd_rz(__fadd_rz(__fadd_rz(__fadd_rz(pfps[index3d], pfu[index4d]), -fpt1), -pfqq[index4d]), pfqq[index4d - SZF]), -pfpw[idz*iLab + idlab]);
            else if(idz == iNz - 1)
                fpt2 = __fadd_rz(__fadd_rz(__fadd_rz(__fadd_rz(pfps[index3d], pfu[index4d]), -fpt1), pfqq[index4d - SZF]), -pfpw[idz*iLab + idlab]);
            else
                ;
            
            /* fpt2 = __fadd_rz(__fadd_rz(pfps[index3d], pfu[index4d]), -fpt1);*/

            pfdv[index4d] = fpt1;

            pfpt[index4d] = MIN(fpt2, pfCt[index4d]);

            /* pfgk[index4d] = fpt1 + pfpt[index4d]; */
        }
    }
}

static __global__ void krnl_qq(float *pfbx, float *pfby, float *pfdv,
                            float *pfps, float *pfpt, float *pfu, 
                            float *pfCt, float *pfgk, float cc,
                            int iNx, int iNy, int iLab, int iNz, int SZF, int szImg, 
                            int blocksInY, double invBlocksInY, float *pfqq, float beta, float *pfpw)
{
   //int blockIdxz = __double2uint_rd(blockIdx.y * invBlocksInY);
   //int blockIdxy = blockIdx.y - __umul24(blockIdxz,blocksInY);
   int blockIdxz = __float2uint_rd(__fdiv_rd(blockIdx.y,blocksInY)); 
   int blockIdxy = blockIdx.y - __umul24(blockIdxz,blocksInY);
   int idx   = __mul24(blockIdx.x,blockDim.x)+threadIdx.x;
   int idy   = __mul24(blockIdxy,blockDim.y)+threadIdx.y;
   int idz   = __mul24(blockIdxz,blockDim.z)+threadIdx.z;
   int index3d, index4d, idlab;
   float fpt1, fpt2, fpt;

    if( idx<iNy && idy<iNx && idz<iNz)
    {
        for (idlab = 0; idlab <iLab; idlab++)
        {
            index3d     = idx +  __umul24(idy, iNy) +  __umul24(idz, SZF);           
            /*index3d           = idx + idy*iNy + (idz+0)*iNx*iNy; */
            /*index3d - SZF     = idx + idy*iNy + (idz-1)*iNx*iNy; */
            /*index3d + SZF     = idx + idy*iNy + (idz+1)*iNx*iNy; */

            //index4d     = idx +  __umul24(idy, iNy) +  __umul24(idz, SZF) + __umul24(idlab, szImg);        
            index4d     = idx + idy*iNy + idz*SZF + idlab*szImg;

            /* index4d          = idx + idy*iNy + (idz + 0)*iNx*iNy + idlab*iNx*iNy*iLab;  */
            /* index4d - SZF    = idx + idy*iNy + (idz - 1)*iNx*iNy + idlab*iNx*iNy*iLab;  */
            /* index4d + SZF    = idx + idy*iNy + (idz + 1)*iNx*iNy + idlab*iNx*iNy*iLab;  */

            if(idz == 0)
                {
                    fpt1 = __fadd_rz(__fadd_rz(__fadd_rz(__fadd_rz(__fadd_rz(pfdv[index4d + SZF], -pfps[index3d + SZF]), pfpt[index4d + SZF]), -pfu[index4d + SZF]), pfqq[index4d + SZF]), pfpw[(idz+1)*iLab + idlab]);
                    /*pfdv[index4d + SZF] - pfps[index3d + SZF] + pfpt[index4d + SZF] - pfu[index4d+SZF] + pfqq[index4d + SZF]; */

                    fpt2 = (__fadd_rz(__fadd_rz(__fadd_rz(pfdv[index4d], - pfps[index3d]), pfpt[index4d]), -pfu[index4d]), pfpw[idz*iLab + idlab]);
                    /*pfdv[index4d] - pfps[index3d] + pfpt[index4d] - pfu[index4d]; */

                    fpt = fpt1 - fpt2;
                    pfqq[index4d] = MAX(MIN(fpt/2, beta), -beta); 
                }
            else if((idz > 0) && (idz < (iNz - 2)))
                {
                    fpt1 = __fadd_rz(__fadd_rz(__fadd_rz(__fadd_rz(__fadd_rz(pfdv[index4d + SZF], -pfps[index3d + SZF]), pfpt[index4d + SZF]), -pfu[index4d + SZF]), pfqq[index4d + SZF]), pfpw[(idz+1)*iLab + idlab]);
                    /*pfdv[index4d + S2D] - pfps[index3d + SZF] + pfpt[index4d + SZF] - pfu[index4d + SZF] + pfqq[index4d + SZF]; */

                    fpt2 = __fadd_rz(__fadd_rz(__fadd_rz(__fadd_rz(__fadd_rz(pfdv[index4d], -pfps[index3d]), pfpt[index4d]), -pfu[index4d]), -pfqq[index4d - SZF]), pfpw[idz*iLab + idlab]);
                    /*pfdv[index4d] - pfps[index1] + pfpt[index4d] - pfu[index4d] -pfqq[index4d-SZF]; */

                    fpt = fpt1 - fpt2;
                    pfqq[index4d] = MAX(MIN(fpt/2, beta), -beta); 
                }
            else if(idz == (iNz - 2))
                {               
                    fpt1 = __fadd_rz(__fadd_rz(__fadd_rz(__fadd_rz(pfdv[index4d + SZF], -pfps[index3d + SZF]), pfpt[index4d + SZF]), -pfu[index4d + SZF]), pfpw[(idz+1)*iLab + idlab]);
                    /*pfdv[index4d + SZF] - pfps[index3d + SZF] + pfpt[index4d + SZF] - pfu[index4d + SZF]; */
    
                    fpt2 = __fadd_rz(__fadd_rz(__fadd_rz(__fadd_rz(__fadd_rz(pfdv[index4d], -pfps[index3d]), pfpt[index4d]), -pfu[index4d]), -pfqq[index4d - SZF]), pfpw[idz*iLab + idlab]);
                    /*pfdv[index4d] - pfps[index3d] + pfpt[index4d] - pfu[index4d] - pfqq[index4d-SZF]; */
        
                    fpt = fpt1 - fpt2;
                    pfqq[index4d] = MAX(MIN(fpt/2, beta), -beta); 
                }
            else if(idz == (iNz - 1))
                {
                /*pfqq[index4d] = 0;*/
                    pfqq[index4d] = 0;
                }
        }
    }
}

static __global__ void krnl_8(float *pfdv, float *pfps, float *pfpt, float *pfu, 
                            float *pfgk, float *FPS, float cc,
                            int iNx, int iNy, int iLab, int iNz, int SZF, int szImg, 
                            int blocksInY, double invBlocksInY, float ft, float *pfqq, float *pfpw)
{

   //int blockIdxz = __double2uint_rd(blockIdx.y * invBlocksInY);
   //int blockIdxy = blockIdx.y - __umul24(blockIdxz,blocksInY);
   int blockIdxz = __float2uint_rd(__fdiv_rd(blockIdx.y,blocksInY)); 
   int blockIdxy = blockIdx.y - __umul24(blockIdxz,blocksInY);
   int idx   = __mul24(blockIdx.x,blockDim.x)+threadIdx.x;
   int idy   = __mul24(blockIdxy,blockDim.y)+threadIdx.y;
   int idz   = __mul24(blockIdxz,blockDim.z)+threadIdx.z;

   int idlab;
   float fpt;
   int index3d, index4d;

   if( idx<iNy && idy<iNx && idz<iNz)
   {
            index3d     = idx +  __umul24(idy, iNy) +  __umul24(idz, SZF);           
            /*index3d           = idx + idy*iNy + (idz+0)*iNx*iNy; */
            /*index3d - SZF     = idx + idy*iNy + (idz-1)*iNx*iNy; */
            /*index3d + SZF     = idx + idy*iNy + (idz+1)*iNx*iNy; */

            fpt = 0;
            for (idlab = 0; idlab < iLab; idlab++) 
                {
                    //index4d     = idx +  __umul24(idy, iNy) +  __umul24(idz, SZF) + __umul24(idlab, szImg);
                    index4d     = idx + idy*iNy + idz*SZF + idlab*szImg;

                    /* index4d          = idx + idy*iNy + (idz + 0)*iNx*iNy + idlab*iNx*iNy*iLab;  */
                    /* index4d - SZF    = idx + idy*iNy + (idz - 1)*iNx*iNy + idlab*iNx*iNy*iLab;  */
                    /* index4d + SZF    = idx + idy*iNy + (idz + 1)*iNx*iNy + idlab*iNx*iNy*iLab;  */
                    
                    /* fpt += __fadd_rz(pfgk[index4d], -pfu[index4d]);*/
                    if(idz == 0)
                        {
                        /* fpt += __fadd_rz(__fadd_rz(pfgk[index4d], -pfu[index4d]), pfqq[index4d]); */
                            pfgk[index4d] = __fadd_rz(__fadd_rz(__fadd_rz(pfdv[index4d], pfpt[index4d]), -pfu[index4d]), pfqq[index4d]);
                            fpt += pfgk[index4d] + pfpw[idz*iLab + idlab];
                        }
                    else if((idz > 0) && (idz < (iNz - 1)))
                        {
                        /*fpt += __fadd_rz(__fadd_rz(__fadd_rz(pfgk[index4d], -pfu[index4d]), pfqq[index4d]), - pfqq[index4d - SZF]);*/
                            pfgk[index4d] = __fadd_rz(__fadd_rz(__fadd_rz(__fadd_rz(pfdv[index4d], pfpt[index4d]), -pfu[index4d]), pfqq[index4d]), -pfqq[index4d - SZF]);
                            fpt += pfgk[index4d] + pfpw[idz*iLab + idlab];
                        }
                    else if(idz == (iNz - 1))    
                        {
                        /*fpt += __fadd_rz(__fadd_rz(pfgk[index4d], -pfu[index4d]), -pfqq[index4d - SZF]);*/
                            pfgk[index4d] = __fadd_rz(__fadd_rz(__fadd_rz(pfdv[index4d], pfpt[index4d]), -pfu[index4d]), -pfqq[index4d - SZF]);
                            fpt += pfgk[index4d] + pfpw[idz*iLab + idlab];
                        }
                    else
                        ;
                }

            pfps[index3d] = __fadd_rz(__fdiv_rz(fpt, iLab), ft);

            for (idlab = 0; idlab < iLab; idlab++) 
                {
                    //index4d     = idx +  __umul24(idy, iNy) +  __umul24(idz, SZF) + __umul24(idlab, szImg);
                    index4d     = idx + idy*iNy + idz*SZF + idlab*szImg;
                    pfgk[index4d] -= pfps[index3d];
                }
    }  
} 

static __global__ void krnl_9(float *pfdv, float *pfps, float *pfpt, float *pfu, 
                            float *pfgk, float *FPS, float cc,
                            int iNx, int iNy, int iLab, int iNz, int SZF, int szImg, 
                            int blocksInY, double invBlocksInY, float ft, float *pfqq, float *pfpw)
{

   //int blockIdxz = __double2uint_rd(blockIdx.y * invBlocksInY);
   //int blockIdxy = blockIdx.y - __umul24(blockIdxz,blocksInY);
   int blockIdxz = __float2uint_rd(__fdiv_rd(blockIdx.y,blocksInY)); 
   int blockIdxy = blockIdx.y - __umul24(blockIdxz,blocksInY);
   int idx   = __mul24(blockIdx.x,blockDim.x)+threadIdx.x;
   int idy   = __mul24(blockIdxy,blockDim.y)+threadIdx.y;
   int idz   = __mul24(blockIdxz,blockDim.z)+threadIdx.z;

   int idlab;
   float fpt;
   int index3d, index4d;

   if( idx<iNy && idy<iNx && idz<iNz)
   {
            index3d     = idx +  __umul24(idy, iNy) +  __umul24(idz, SZF);           
            /*index3d           = idx + idy*iNy + (idz+0)*iNx*iNy; */
            /*index3d - SZF     = idx + idy*iNy + (idz-1)*iNx*iNy; */
            /*index3d + SZF     = idx + idy*iNy + (idz+1)*iNx*iNy; */
            FPS[index3d] = 0;

            for (idlab = 0; idlab < iLab; idlab++) 
                {
                    //index4d     = idx +  __umul24(idy, iNy) +  __umul24(idz, SZF) + __umul24(idlab, szImg);
                    index4d     = idx + idy*iNy + idz*SZF + idlab*szImg;

                    /* fpt = __fadd_rz(pfgk[index4d - SZF], -pfps[index3d - SZF]);*/
                    if(idz == 0)
                        /* fpt = __fadd_rz(__fadd_rz(pfgk[index4d], -pfps[index3d]), pfqq[index4d]);  */
                        fpt = __fadd_rz(__fadd_rz(__fadd_rz(__fadd_rz(pfdv[index4d], -pfps[index3d]), pfpt[index4d]), pfqq[index4d]), pfpw[idz*iLab + idlab]);
                    else if((idz > 0) && (idz < (iNz - 1)))
                        /* fpt = __fadd_rz(__fadd_rz(__fadd_rz(pfgk[index4d], -pfps[index3d]), pfqq[index4d]), -pfqq[index4d - SZF]); */
                        fpt = __fadd_rz(__fadd_rz(__fadd_rz(__fadd_rz(__fadd_rz(pfdv[index4d], -pfps[index3d]), pfpt[index4d]), pfqq[index4d]), -pfqq[index4d - SZF]), pfpw[idz*iLab + idlab]);
                    else if(idz == (iNz - 1))    
                        /* fpt = __fadd_rz(__fadd_rz(pfgk[index4d], -pfps[index3d]), -pfqq[index4d - SZF]); */
                        fpt = __fadd_rz(__fadd_rz(__fadd_rz(__fadd_rz(pfdv[index4d], -pfps[index3d]), pfpt[index4d]), -pfqq[index4d - SZF]), pfpw[idz*iLab + idlab]);
                    else
                        ;
                    fpt = __fmul_rz(fpt,cc);
                    pfu[index4d] -= fpt;
                    FPS[index3d] += ABS(fpt);
                }
    }  
} 