function FFT(FD_lung_reg, RespData, FFT_VDP)

% read  the registered H MR images.
reg_lung = metaImageRead(FD_lung_reg);
[rows, cols, slices] = size(reg_lung);

Freq_ventilation = zeros(rows, cols);

%% Resp Input
% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

time_bellow = 0.04;                % time intervals for respiratory bellow signals
time_MRI = 0.25;                   % time intervals for MRI slices
adjust = time_MRI/time_bellow;     % how many bellow points correspond to one MRI slice
M1 = textread(RespData);
len = size(M1, 1);                 % total length of respiratory bellow data (gabaga in the front needs to be removed)

data = [];                         % bellow data that corresponds to the MRI slices
adj = int32(slices*adjust);        % back-tracking to find the bellow data - with a lenght of "adj"

for z = (len - adj + 1):len        % changed from "len - adj" to "len- adj + 1" by Fumin, Oct 4, 2016
    data(z - (len - adj + 1) + 1) = M1(z);
end

AA = fft(data);                    % fft of bellow data with gaussian smoothing    
nfft = length(AA);
AA = AA(1:nfft/2);
AA_mag_half1 = abs(AA);
Fs = 1/time_bellow;
B1 = (0:nfft/2-1)*Fs/nfft;
xc = B1;
H = fspecial('gaussian', 5, 2);
yc = imfilter(AA_mag_half1,H,'replicate');

% find the respiratory frequency.
[pks, locs] = findpeaks(yc,'MinPeakDistance',6);
[pk, i] = max(pks);
f_input = xc(locs(i));

% the slice number of FFT_map at the bellow respiratory frequency
Fs_image = 1/time_MRI;
select_slice_freq = round(f_input*slices/Fs_image);

for i = 1 : rows
    for j = 1: cols
%         disp(['FFT of line -- ' num2str((i-1)*cols+j) ' -- out of -- ' num2str(rows*cols) ' now!!'])
        tem = reg_lung(i,j,:);
        Hann_window = hann(slices);
        tem_han = tem(:).*Hann_window;
        AA_image = fft(tem_han);
        nfft_image = length(AA_image);
        AAA_image = AA_image(1:nfft_image/2);
        AA_mag_half1_image = abs(AAA_image);
%         H_image = fspecial('gaussian', 5, 2);
        Freq_temp = imfilter(AA_mag_half1_image,H,'replicate');
        FFT_map(i,j,:) = Freq_temp/max(Freq_temp(:));
    end
end

% FFT_map outside the lung is useless, and set to 0.
FFT_map(isnan(FFT_map)) = 0;

n_sig = 1.0;
for i = 1:size(FFT_map,3)
    temtem = FFT_map(:,:,i);                        % get the mean and std of FFT values within the lung in the current slice
    mn = mean(temtem(:));                           % mean and std of FFT values within hte lung
    sig = std(temtem(:));
    
    % squeeze FFT values to the range [mean - sig : mean + sig]. Note that the
    % original 0s outside the lung will also be changed.
    FFT_map(:,:,i) = max(min(FFT_map(:,:,i), mn + n_sig*sig), mn - n_sig*sig);

    temtem = FFT_map(:,:,i);
    imgMin = min(temtem(:));
    imgMax = max(temtem(:));
    newRange = 100;
    FFT_map(:,:,i) = (FFT_map(:,:,i) - imgMin) / (imgMax - imgMin) * newRange;
    
    H = fspecial('gaussian', 4, 1);
    FFT_map(:,:,i) = imfilter(FFT_map(:,:,i),H,'replicate');       
end


FD_map = FFT_map(:,:,select_slice_freq);
save(FFT_VDP, 'FFT_map', 'FD_map')
end
